--------------------------------------------------------------------------------
-- Company: 
-- Engineer:
--
-- Create Date:    11:39:18 01/27/06
-- Design Name:    
-- Module Name:    HTR_LSB - Behavioral
-- Project Name:   
-- Target Device:  
-- Tool versions:  
-- Description:
--
-- Dependencies:
-- 
-- Revision:
-- Revision 2 - File Created
-- Additional Comments:
-- 
-- read
-- address 0x0		Tp_x to GOLx_data mapping scheme number
-- address 0x4		csr
-- address 0x8		GOL_ready & QPLL_locked & QPLL_err & bc_err & neg_edge & run
-- address 0xc		i2c_busy & i2c_fail & 	"00" & locked & ttc_status(1) & rx_status(1)
-- address 0x10	GOL0 spy data
-- address 0x14	GOL1 spy data
-- address 0x18	GOL2 spy data
-- address 0x1c	i2c_data
-- address 0x20	not used
-- address 0x24	ps state low 7 bits
-- address 0x28	ps state high 7 bits
-- address 0x2c	firmware version
-- address 0x30	GOL0 spy data address counter 8 LSBs
-- address 0x34	GOL1 spy data address counter 8 LSBs
-- address 0x38	GOL2 spy data address counter 8 LSBs
-- address 0x3c	GOLx spy data address counter 2 MSBs
-- write
-- address 0x0		Tp_x to GOLx_data mapping scheme number		
-- address 0x4		csr		
-- address 0xc		i2c address register		
-- address 0x1c	i2c data register		
-- address 0x30	GOL0 spy data address counter bit 9 downto 2 and resets bit 1 and 0
-- address 0x34	GOL1 spy data address counter bit 9 downto 2 and resets bit 1 and 0
-- address 0x38	GOL2 spy data address counter bit 9 downto 2 and resets bit 1 and 0
-- address 0x40	commands
--						bit0 clears FLASH buffer address counter
--						bit1 takes a snapshot of GOLx_data		
-- address 0x44	FLASH buffer data register		
-- address 0x48	resets		
-- address 0x4c	csr1		
-- address 0x50	test_data(7:0)		
-- address 0x54	test_data(15:8)		
-- address 0x58	test_data(23:16)		
-- address 0x5c	test_data(31:24)		
--------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

---- Uncomment the following library declaration if instantiating
---- any Xilinx primitives in this code.
library UNISIM;
use UNISIM.VComponents.all;

entity OSLB is
    Port ( TP_a : in std_logic_vector(35 downto 0);
           TP_b : in std_logic_vector(35 downto 0);
           loc_bus : inout std_logic_vector(7 downto 0);
           loc_add : in std_logic_vector(6 downto 0);
           loc_rw : in std_logic;
           loc_cs : in std_logic;
           loc_ctrl : out std_logic_vector(2 downto 0);
           init_b : out std_logic;
-- loc_spare signals are unused and grounded by CPLD
           loc_spare : in std_logic_vector(7 downto 0);
           ttc_brcst : in std_logic_vector(7 downto 2);
           ttc_brcst_strb : in std_logic;
           ttc_clk_des : in std_logic;
           ttc_bcr : in std_logic;
           rx_bc0_p : in std_logic;
           rx_bc0_n : in std_logic;
           rx_clk_p : in std_logic;
           rx_clk_n : in std_logic;
           GOL_ready : in std_logic_vector(2 downto 0);
           GOL0_data : out std_logic_vector(31 downto 0);
           GOL1_data : out std_logic_vector(31 downto 0);
           GOL2_data : out std_logic_vector(31 downto 0);
           GOL0_flag : out std_logic_vector(1 downto 0);
           GOL1_flag : out std_logic_vector(1 downto 0);
           GOL2_flag : out std_logic_vector(1 downto 0);
           GOL_reset_b : out std_logic_vector(2 downto 0);
           GOL_negedge : out std_logic;
           GOL_cav : out std_logic_vector(2 downto 0);
           GOL_dav : out std_logic_vector(2 downto 0);
           GOL_FF : out std_logic_vector(2 downto 0);
           GOL_SCL : out std_logic_vector(2 downto 0);
           GOL_SDA : inout std_logic;
           QPLL_locked : in std_logic;
           QPLL_err : in std_logic;
           QPLL_reset_b : out std_logic;
           CB_SI : out std_logic;
           CB_EN : out std_logic;
           CB_CK : out std_logic;
           GLINK : out std_logic
			  );
end OSLB;

architecture Behavioral of OSLB is
constant version : std_logic_vector(7 downto 0) := x"82";
constant sel_flash : std_logic_vector(6 downto 5) := "01";
constant sel_maddr : std_logic_vector(6 downto 2) := "11000";
COMPONENT m32_1
	PORT(
		d : IN std_logic_vector(31 downto 0);
		s : IN std_logic_vector(4 downto 0);          
		o : OUT std_logic
		);
END COMPONENT;
COMPONENT m16_1
	PORT(
		d : IN std_logic_vector(15 downto 0);
		s : IN std_logic_vector(3 downto 0);          
		o : OUT std_logic
		);
END COMPONENT;
COMPONENT M16_1b
	PORT(
		d0 : IN std_logic;
		d1 : IN std_logic;
		d2 : IN std_logic;
		d3 : IN std_logic;
		d4 : IN std_logic;
		d5 : IN std_logic;
		d6 : IN std_logic;
		d7 : IN std_logic;
		d8 : IN std_logic;
		d9 : IN std_logic;
		d10 : IN std_logic;
		d11 : IN std_logic;
		d12 : IN std_logic;
		d13 : IN std_logic;
		d14 : IN std_logic;
		d15 : IN std_logic;
		s : IN std_logic_vector(3 downto 0);          
		o : OUT std_logic
		);
END COMPONENT;
component xnor8
    port (o : out std_logic; i0 : in std_logic; i1 : in std_logic;
	       i2 : in std_logic; i3 : in std_logic; i4 : in std_logic;
	       i5 : in std_logic; i6 : in std_logic; i7 : in std_logic
	 );
end component;
COMPONENT rpc_sender
  generic (
    LPM_PART_WIDTH		:positive := 8;
    LPM_PART_NUM		:positive := 4;
    LPM_CHECK_WIDTH		:positive := 8;
    INPUT_REGISTERED		:boolean := true;
    OUTPUT_REGISTERED		:boolean := true
  );
	PORT(
		resetN : IN std_logic;
		clock : IN std_logic;
		part_ena : IN std_logic;
		check_ena : IN std_logic;
		check_data : IN std_logic_vector(7 downto 0);
		check_data_ena : IN std_logic;
		in_data : IN std_logic_vector(23 downto 0);
		test_data : IN std_logic_vector(31 downto 0);
		test_ena : IN std_logic;
		test_rand_ena : IN std_logic;          
		out_data : OUT std_logic_vector(31 downto 0)
		);
END COMPONENT;
COMPONENT ram256x8
	PORT(
		wclk : IN std_logic;
		rclk : IN std_logic;
		we : IN std_logic;
		re : IN std_logic;
		di : IN std_logic_vector(7 downto 0);
		wa : IN std_logic_vector(7 downto 0);
		ra : IN std_logic_vector(7 downto 0);          
		do : OUT std_logic_vector(7 downto 0)
		);
END COMPONENT;
signal rx_status : std_logic_vector(7 downto 0) := (others => '0');
signal ttc_status : std_logic_vector(7 downto 0) := (others => '0');
signal locked : std_logic_vector(1 downto 0) := (others => '0');
signal rst_cntr : std_logic_vector(3 downto 0) := x"0";
signal powerup : std_logic := '1';
signal locked_q : std_logic_vector(1 downto 0) := (others => '0');
signal startup : std_logic := '0';
signal ttc_clk : std_logic := '0';
signal ttc_clk_dcm : std_logic := '0';
signal ttc_clk2x : std_logic := '0';
signal ttc_clk2x_dcm : std_logic := '0';
signal ttc_cmd : std_logic_vector(5 downto 0) := "000000";
signal run : std_logic := '0';
signal rx_clk_in : std_logic := '0';
signal rx_clk : std_logic := '0';
signal rx_clk_dcm : std_logic := '0';
signal rx_clk90 : std_logic := '0';
signal rx_clk90_dcm : std_logic := '0';
signal rx_bc0 : std_logic := '0';
signal got_bc0 : std_logic := '0';
signal bc_err : std_logic := '0';
signal bc_errp : std_logic := '0';
signal s : std_logic_vector(431 downto 0) := (others => '0');
signal data : std_logic_vector(71 downto 0) := (others => '0');
signal data_a : std_logic_vector(71 downto 0) := (others => '0');
signal data_b : std_logic_vector(71 downto 0) := (others => '0');
signal data_ba : std_logic_vector(71 downto 0) := (others => '0');
signal data_bb : std_logic_vector(71 downto 0) := (others => '0');
signal mux_in_a : std_logic_vector(31 downto 0) := (others => '0');
signal mux_in_b : std_logic_vector(15 downto 0) := (others => '0');
signal GOL0_data_i : std_logic_vector(31 downto 0) := (others => '0');
signal GOL1_data_i : std_logic_vector(31 downto 0) := (others => '0');
signal GOL2_data_i : std_logic_vector(31 downto 0) := (others => '0');
signal check_data : std_logic_vector(7 downto 0) := (others => '0');
signal lut_addr : std_logic_vector(8 downto 0) := (others => '0');
signal lut_addrb : std_logic_vector(8 downto 0) := (others => '0');
signal r_data : std_logic_vector(7 downto 0) := (others => '0');
signal csr : std_logic_vector(7 downto 0) := x"00";
signal csr1 : std_logic_vector(7 downto 0) := x"00";
signal resets : std_logic_vector(7 downto 0) := x"00";
signal start_reset :std_logic := '0';
signal cmd : std_logic_vector(7 downto 0) := x"00";
signal rst_dl :std_logic_vector(15 downto 0) := x"0000";
signal test_data : std_logic_vector(31 downto 0) := x"00000000";
signal test_datap : std_logic_vector(31 downto 0) := x"00000000";
signal rst_dcm :std_logic := '0';
signal psen :std_logic := '0';
signal psdone :std_logic := '0';
signal got_psdone :std_logic := '0';
signal psinc :std_logic := '0';
signal neg_edge :std_logic := '0';
signal calib_going :std_logic := '0';
signal ttc_clk_des_q :std_logic_vector(1 downto 0) := (others => '0');
signal ttc_clk_des_q2 :std_logic_vector(1 downto 0) := (others => '0');
signal ps :std_logic_vector(6 downto 0) := (others => '0');
signal ps_state : std_logic_vector(10 downto 0) := (others => '0');
signal ps_cntr :std_logic_vector(5 downto 0) := (others => '0');
signal addr :std_logic_vector(13 downto 0) := (others => '0');
signal loc_rw_q : std_logic := '0';
signal loc_clk : std_logic := '0';
signal flash_clk : std_logic := '0';
signal loc_clkn : std_logic := '0';
signal soft_sync : std_logic_vector(2 downto 0) := (others => '0');
signal sync_active : std_logic_vector(1 downto 0) := (others => '0');
signal sync_active0_dl : std_logic := '0';
signal sync_time : std_logic_vector(15 downto 0) := (others => '0');
signal sync_idle : std_logic_vector(15 downto 0) := (others => '0');
signal sync_test : std_logic_vector(15 downto 0) := (others => '0');
signal sync_cntr : std_logic_vector(7 downto 0) := (others => '0');
signal rst_sync_cntr : std_logic_vector(2 downto 0) := (others => '0');
signal sync_cmd : std_logic_vector(7 downto 2) := (others => '0');
signal sync_mask : std_logic_vector(7 downto 2) := (others => '0');
signal i2c_sync : std_logic_vector(1 downto 0) := (others => '0');
signal i2c_addr : std_logic_vector(4 downto 0) := (others => '0');
signal i2c_data : std_logic_vector(7 downto 0) := (others => '0');
signal i2c_wdata : std_logic_vector(7 downto 0) := (others => '0');
signal i2c_sr : std_logic_vector(17 downto 0) := (others => '0');
signal i2c_phase : std_logic_vector(2 downto 0) := (others => '0');
signal i2c_scl : std_logic_vector(2 downto 0) := (others => '0');
signal i2c_cntr : std_logic_vector(4 downto 0) := (others => '0');
signal clk_div : std_logic_vector(4 downto 0) := (others => '0');
signal i2c_ec : std_logic := '0';
signal i2c_read : std_logic := '0';
signal i2c_eoc : std_logic := '0';
signal data_cycle : std_logic := '0';
signal i2c_toggle : std_logic := '0';
signal i2c_start : std_logic := '0';
signal i2c_active : std_logic := '0';
signal i2c_done : std_logic := '0';
signal i2c_fail : std_logic := '0';
signal i2c_busy : std_logic := '0';
signal we_flash : std_logic := '0';
signal empty_buffer :std_logic := '1';
signal flash_addra : std_logic_vector(10 downto 0) := "00000000000";
signal flash_addrb : std_logic_vector(10 downto 0) := "00000000000";
signal flash_buf : std_logic_vector(7 downto 0) := (others => '0');
signal divider : std_logic_vector(4 downto 0) := "00000";
signal ce_flash_addrb : std_logic := '0';
signal flash_ce : std_logic := '0';
signal rd_buffer : std_logic := '0';
signal rd_done : std_logic_vector(1 downto 0) := (others => '0');
signal clk_ctrl : std_logic_vector(10 downto 0) := (others => '0');
signal en_CB_CK : std_logic_vector(11 downto 0) := (others => '0');
signal ttc_clkn : std_logic := '0';
signal cmd1_sync : std_logic_vector(1 downto 0) := (others => '0');
signal we_gol : std_logic := '0';
signal we_gol2 : std_logic := '0';
signal sel_gol1 : std_logic := '0';
signal gol_spy_data : std_logic_vector(7 downto 0) := (others => '0');
signal gol_spy_addr : std_logic_vector(15 downto 0) := (others => '0');
signal gol_addrb : std_logic_vector(8 downto 0) := (others => '0');
signal gol2_addrb : std_logic_vector(8 downto 0) := (others => '0');
signal gol_addra : std_logic_vector(10 downto 0) := (others => '0');
signal gol0_addra : std_logic_vector(10 downto 0) := (others => '0');
signal gol1_addra : std_logic_vector(10 downto 0) := (others => '0');
signal GOL_datap : std_logic_vector(31 downto 0) := (others => '0');
signal gol_buf : std_logic_vector(7 downto 0) := (others => '0');
signal gol2_buf : std_logic_vector(7 downto 0) := (others => '0');
signal status0 : std_logic_vector(7 downto 0) := (others => '0');
signal status1 : std_logic_vector(7 downto 0) := (others => '0');
signal oslb_reg : std_logic_vector(7 downto 0) := (others => '0');
signal reset_senderN : std_logic;
signal check_ena : std_logic;
signal check_data_ena : std_logic;
signal test_ena : std_logic;
signal test_rand_ena : std_logic;
signal DCM_lock_sync : std_logic_vector(1 downto 0) := (others => '0');
signal gol0_ready_sync : std_logic_vector(1 downto 0) := (others => '0');
signal gol1_ready_sync : std_logic_vector(1 downto 0) := (others => '0');
signal gol2_ready_sync : std_logic_vector(1 downto 0) := (others => '0');
signal qpll_lock_sync : std_logic_vector(1 downto 0) := (others => '0');
signal qpll_err_sync : std_logic_vector(1 downto 0) := (others => '0');
signal qpll_err_toggle : std_logic := '0';
signal err : std_logic := '0';
signal bc0_err : std_logic := '0';
signal bc0 : std_logic := '0';
signal bc0_a : std_logic;
signal bc0_b : std_logic;
signal bcn_mismatch : std_logic := '0';
signal err_cntr : std_logic_vector(7 downto 0) := (others => '0');
signal err_cntrp : std_logic_vector(7 downto 0) := (others => '0');
signal d_err_cntr : std_logic_vector(7 downto 0) := (others => '0');
signal err_cntr_wa : std_logic_vector(2 downto 0) := (others => '0');
signal err_sel : std_logic_vector(7 downto 0) := (others => '0');
signal we_err_cntr : std_logic;
signal bcn : std_logic_vector(11 downto 0) := (others => '0');
signal hist_sel : std_logic_vector(7 downto 0) := (others => '0');
signal TP_sel_in : std_logic_vector(63 downto 0) := (others => '0');
signal TP_sel : std_logic_vector(3 downto 0) := (others => '0');
signal TP_sel_q : std_logic_vector(7 downto 0) := (others => '0');
signal sum : std_logic_vector(71 downto 0) := (others => '0');
signal hist : std_logic_vector(63 downto 0) := (others => '0');
signal hist_ra : std_logic_vector(15 downto 0) := (others => '0');
signal hist_do : std_logic_vector(63 downto 0) := (others => '0');
signal hist_data : std_logic_vector(7 downto 0) := (others => '0');
signal we_hist : std_logic_vector(7 downto 0) := (others => '0');
signal rst_hist : std_logic_vector(3 downto 0) := (others => '0');
signal sim_hist : std_logic_vector(3 downto 0) := (others => '0');
signal orbit_cntr : std_logic_vector(3 downto 0) := (others => '0');
signal bcnp : std_logic_vector(3 downto 0) := (others => '0');
signal sim_hit : std_logic_vector(3 downto 0) := (others => '0');
attribute iostandard : string;
attribute iostandard of i_rxclk : label is "lvds_25";
attribute iostandard of i_rxbc0 : label is "lvds_25";
attribute iob : string;
attribute iob of ttc_clk_des_q : signal is "true";
attribute iobdelay : string;
attribute iobdelay of TP_a : signal is "IFD";
attribute iobdelay of TP_b : signal is "IFD";
attribute init : string;
attribute init of rd_buffer : signal is "0";
begin
GLINK <= '0';
init_b <= '1';
flash_clk <= loc_spare(3);
loc_clk <= loc_spare(5);
gol0_data <= gol0_data_i;
gol1_data <= gol1_data_i;
gol2_data <= gol2_data_i;
process(ttc_clk_des,ttc_status)
begin
-- if clkin stops toggling, reset DCM until clkin comes back
  if(ttc_status(1) = '1' or rx_status(1) = '1')then
    rst_cntr <= x"8";
  elsif(ttc_clk_des'event and ttc_clk_des = '1')then
    if(rst_cntr(3) = '1')then
      rst_cntr <= rst_cntr + 1;
	 end if;
  end if;
end process;
i_dcm_ttc: dcm
	port map	
         ( CLKIN   => ttc_clk_des ,
           CLKFB   => ttc_clk ,
           RST     => rst_cntr(3) ,
           DSSEN   => '0' ,
           PSINCDEC=> '0' ,
           PSEN    => '0' ,
           PSCLK   => '0' ,
			  CLK0    => ttc_clk_dcm ,
			  CLK2x    => ttc_clk2x_dcm ,
			  STATUS  => ttc_status ,
			  LOCKED  => locked(0));
bufg_clk: bufg port map(o => ttc_clk, i => ttc_clk_dcm);
bufg_clk2x: bufg port map(o => ttc_clk2x, i => ttc_clk2x_dcm);
process(ttc_clk,locked)
begin
  if(locked(0) = '0')then
    powerup <= '1';
	 locked_q <= "00";
	 startup <= '0';
  elsif(ttc_clk'event and ttc_clk = '1')then
	 locked_q <= locked_q(0) & locked(1);
	 if(locked_q(1) = '1' and powerup = '1')then
	   startup <= '1';
	 else
	   startup <= '0';
	 end if;
	 if(startup = '1')then
	   powerup <= '0';
	 end if;
  end if;
end process;
-- PCK2111 control
process(ttc_clk)
begin
  if(ttc_clk'event and ttc_clk = '1')then
    if(resets(6) = '1')then
	   clk_ctrl <= "000000" & not csr(2) & '1' & not csr(0) & not csr(1) & csr(3);
	 else
		clk_ctrl <= clk_ctrl(9 downto 0) & '0';
	 end if;
	 en_CB_CK <= en_CB_CK(10 downto 0) & resets(6);
  end if;
end process;
CB_EN <= not resets(6);
CB_SI <= clk_ctrl(10);
ttc_clkn <= not ttc_clk;
i_CB_CK:  fddrrse
	      port map(
			    d0 => '0',
				 d1 => en_CB_CK(11),
				 c0 => ttc_clk,
				 c1 => ttc_clkn,
				 ce => '1',
				 r  => '0',
				 s  => '0',
				 q  => CB_CK);
i_rxclk: ibufgds port map(i => rx_clk_p, ib => rx_clk_n, o => rx_clk_in);
i_rxbc0: ibufds port map(i => rx_bc0_p, ib => rx_bc0_n, o => rx_bc0);
process(rx_clk)
begin
  if(rx_clk'event and rx_clk = '1')then
    if(rx_bc0 = '1')then
	   got_bc0 <= '1';
	 else
	   got_bc0 <= '0';
	 end if;
  end if;
end process;
-- mapping control ROMs, init values are in ucf file
i_g_rom: for i in 0 to 5 generate
  i_rom: RAMB16_S36_S36
		   port map(
			   dia => x"00000000",
				dib => x"00000000",
				dipa => x"0",
				dipb => x"0",
				ena => '1',
				enb => '1',
				wea => '0',
				web => '0',
				ssra => '0',
				ssrb => '0',
				clka => ttc_clk,
				clkb => ttc_clk,
				addra => lut_addr,
				addrb => lut_addrb,
				doa => s(31+i*72 downto i*72),
				dopa => s(35+i*72 downto 32+i*72),
				dob => s(67+i*72 downto 36+i*72),
				dopb => s(71+i*72 downto 68+i*72)
				);
end generate;
lut_addrb <= lut_addr(8 downto 1) & '1';
-- there are 24 * 3 mux outputs for GOL0, GOL1 and GOL2 (bits 23:0)
-- each output needs 6 bits s to select from tp_a(23:0) and tp_b(23:0),
-- for test purposes, test_data(23:0) and tp_a(35:24) and tp_b(35:24) can also be selected
-- tp_a(23:0) is selected when s = 23 to 0, while tp_b(23:0) is selectd when s = 47 to 24
-- if csr(4) = '1', tp_a(23:0) and tp_b(23:0) are replaced by test_data(23:0)
-- if s = 48 then output (17:6) are tp_a(35:24)
-- if s = 49 then output (17:6) are tp_b(35:24) 
-- multiplexers
i_g0: for i in 0 to 71 generate
    i_mux_a: m32_1 PORT MAP(
									d =>  mux_in_a,
									s => s(4+i*6 downto i*6),
									o => data_a(i));
    i_mux_ba: m16_1 PORT MAP(
									d => mux_in_b,
									s => s(3+i*6 downto i*6),
									o => data_b(i));
--									o => data_ba(i));
--	 i_muxF8: muxF8 port map(
--									i0 => data_ba(i),
--									i1 => data_bb(i),
--									s => s(4+i*6),
--									o => data_b(i));	
--	data(i) <= (not s(5+i*6) and data_a(i)) or (s(5+i*6) and data_b(i));	
	data(i) <= (not s(5+i*6) and data_a(i)) or (s(5+i*6) and not s(4+i*6) and data_b(i));	
end generate;
mux_in_a <= TP_b(7 downto 0) & TP_a(23 downto 0) when csr(4) = '0' else 
				not test_data(7 downto 0) & test_data(23 downto 0);
mux_in_b <= TP_b(23 downto 8) when csr(4) = '0' else not test_data(23 downto 8);
--mux_in_a <= TP_b(7 downto 0) & TP_a(23 downto 0); 
--mux_in_b <= TP_b(23 downto 8);
--i_g1: for i in 0 to 2 generate
--  data_bb(5+i*19 downto i*19) <= "000000";
--  data_bb(18+i*19) <= '0';
--end generate;
--process(lut_addr,TP_a,TP_b)
--begin
--  for i in 0 to 11 loop
--    if(lut_addr(8) = '0')then
--		data_bb(6+i)  <= '0';
--		data_bb(25+i) <= '0';
--		data_bb(44+i) <= '0';
--	 elsif(lut_addr(2) = '0')then
--		data_bb(6+i)  <= TP_a(24+i);
--		data_bb(25+i) <= TP_a(24+i);
--		data_bb(44+i) <= TP_a(24+i);
--	 else	
--		data_bb(6+i)  <= TP_b(24+i);
--		data_bb(25+i) <= TP_b(24+i);
--		data_bb(44+i) <= TP_b(24+i);
--    end if;	 
--  end loop;  
--end process;
GOL_cav <= "000";
process(ttc_clk)
begin
  if(ttc_clk'event and ttc_clk = '1')then
	  if(sync_active0_dl = '0' and (run = '1' or sync_active(1) = '1'))then
      GOL_dav <= "111";
	  else
      GOL_dav <= "000";
    end if;
  end if;
end process;
GOL0_flag <= "00";
GOL1_flag <= "00";
GOL2_flag <= "00";
GOL_FF <= "000";
GOL_reset_b <= not resets(5 downto 3);
QPLL_reset_b <= not resets(2);
run <= csr(7);
process(ttc_clk)
begin
  if(ttc_clk'event and ttc_clk = '1')then
	 soft_sync(2 downto 1) <= soft_sync(1 downto 0);
	 rst_sync_cntr(2 downto 1) <= rst_sync_cntr(1 downto 0);
	 if(rst_sync_cntr(2) /= rst_sync_cntr(1))then
	   sync_cntr <= x"00";
    elsif(((ttc_brcst_strb = '1' and sync_mask /= "111111" and (ttc_brcst and not sync_mask) = (sync_cmd and not sync_mask)) or
			soft_sync(2) /= soft_sync(1)) and csr(5) = '1' and sync_active = "00" and sync_cntr /= x"ff")then
	   sync_cntr <= sync_cntr + 1;
	 end if;
    if(((ttc_brcst_strb = '1' and sync_mask /= "111111" and (ttc_brcst and not sync_mask) = (sync_cmd and not sync_mask)) or
			soft_sync(2) /= soft_sync(1)) and csr(5) = '1' and sync_active = "00")then
	   sync_active(0) <= '1';
	 elsif(sync_time = sync_idle)then
	   sync_active(0) <= '0';
	 end if;
	 sync_active0_dl <= sync_active(0);
	 if(sync_active(0) = '1' and sync_time = sync_idle)then
	   sync_active(1) <= '1';
	 elsif(sync_time = sync_test)then
	   sync_active(1) <= '0';
	 end if;
	 if(sync_active = "00" or (sync_active(0) = '1' and sync_time = sync_idle))then
	   sync_time <= (others => '0');
	 else
	   sync_time <= sync_time + 1;
	 end if;
  end if;
end process;
status0 <= GOL_ready & QPLL_locked & QPLL_err & bc_err & neg_edge & run;
status1 <= i2c_busy & i2c_fail & 	"00" & locked & ttc_status(1) & rx_status(1);
r_data <= flash_buf when rd_buffer = '1' else oslb_reg;
process(loc_add,lut_addr,csr,status0,status1,gol_spy_data,gol_spy_addr,i2c_data,
		  sync_cntr,ps_state,sync_idle,sync_test,csr1,test_data,err_cntr,err_sel,sync_cmd,sync_mask,
		  hist_data,hist_sel,hist_ra)
begin
  case loc_add(4 downto 0) is
    when "00000" => oslb_reg <= lut_addr(8 downto 1);
    when "00001" => oslb_reg <= csr;
    when "00010" => oslb_reg <= status0;
    when "00011" => oslb_reg <= status1;
    when "00100" => oslb_reg <= gol_spy_data;
    when "00101" => oslb_reg <= gol_spy_addr(7 downto 0);
    when "00110" => oslb_reg <= gol_spy_addr(15 downto 8);
    when "00111" => oslb_reg <= i2c_data;
    when "01000" => oslb_reg <= sync_cntr;
    when "01001" => oslb_reg <= '0' & ps_state(6 downto 0);
    when "01010" => oslb_reg <= "0000" & ps_state(10 downto 7);
    when "01011" => oslb_reg <= version;
    when "01100" => oslb_reg <= sync_idle(7 downto 0);
    when "01101" => oslb_reg <= sync_idle(15 downto 8);
    when "01110" => oslb_reg <= sync_test(7 downto 0);
    when "01111" => oslb_reg <= sync_test(15 downto 8);
    when "10011" => oslb_reg <= csr1;
    when "10100" => oslb_reg <= test_data(7 downto 0);
    when "10101" => oslb_reg <= test_data(15 downto 8);
    when "10110" => oslb_reg <= test_data(23 downto 16);
    when "10111" => oslb_reg <= test_data(31 downto 24);
    when "11000" => oslb_reg <= err_cntr;
    when "11001" => oslb_reg <= err_sel;
    when "11010" => oslb_reg <= sync_cmd & "00";
    when "11011" => oslb_reg <= sync_mask & "00";
    when "11100" => oslb_reg <= hist_data;
    when "11101" => oslb_reg <= hist_sel;
    when "11110" => oslb_reg <= hist_ra(7 downto 0);
    when "11111" => oslb_reg <= hist_ra(15 downto 8);
    when others => oslb_reg <= x"00";
  end case;
end process;
loc_bus <= r_data when (loc_cs = '0' and loc_rw = '1' and loc_add(6 downto 5) /= sel_flash) or rd_buffer = '1' else (others => 'Z');
process(loc_clk)
begin
  if(loc_clk'event and loc_clk = '0')then
    loc_rw_q <= loc_rw;
  end if;
end process;
process(loc_clk)
begin
  if(loc_clk'event and loc_clk = '1')then
    if(loc_rw_q = '0' and	loc_add = "0000000")then
	   lut_addr <= loc_bus & '0';
	 end if;
    if(loc_rw_q = '0' and	loc_add = "0000001")then
	   csr <= loc_bus;
	 end if;
    if(loc_rw_q = '0')then
      if(loc_add = "0000101")then
	     gol_spy_addr(15 downto 8) <= x"0" & loc_bus(3 downto 0);
		end if;
      if(loc_add = "0000110")then
	     gol_spy_addr(7 downto 0) <= loc_bus;
		end if;
	 elsif(loc_add = "0000100")then
	   if(gol_spy_addr(11 downto 0) = x"bff")then
	     gol_spy_addr(11 downto 0) <= x"000";
		else
	     gol_spy_addr(11 downto 0) <= gol_spy_addr(11 downto 0) + 1;
		end if;
	 end if;
    if(loc_rw_q = '0' and	loc_add = "0001100")then
	   sync_idle(7 downto 0) <= loc_bus;
	 end if;
    if(loc_rw_q = '0' and	loc_add = "0001101")then
	   sync_idle(15 downto 8) <= loc_bus;
	 end if;
    if(loc_rw_q = '0' and	loc_add = "0001110")then
	   sync_test(7 downto 0) <= loc_bus;
	 end if;
    if(loc_rw_q = '0' and	loc_add = "0001111")then
	   sync_test(15 downto 8) <= loc_bus;
	 end if;
    if(loc_rw_q = '0' and	loc_add = "0010000")then
	   rst_hist(0) <= rst_hist(0) xor loc_bus(2);
	   sim_hist(0) <= sim_hist(0) xor (loc_bus(3) and hist_sel(3) and hist_sel(2));
	   soft_sync(0) <= soft_sync(0) xor loc_bus(4);
	   rst_sync_cntr(0) <= rst_sync_cntr(0) xor loc_bus(5);
	 end if;
    if(loc_rw_q = '0' and	loc_add = "0010011")then
	   csr1 <= loc_bus;
	 end if;
    if(loc_rw_q = '0' and	loc_add = "0010100")then
	   test_data(7 downto 0) <= loc_bus;
	 end if;
    if(loc_rw_q = '0' and	loc_add = "0010101")then
	   test_data(15 downto 8) <= loc_bus;
	 end if;
    if(loc_rw_q = '0' and	loc_add = "0010110")then
	   test_data(23 downto 16) <= loc_bus;
	 end if;
    if(loc_rw_q = '0' and	loc_add = "0010111")then
	   test_data(31 downto 24) <= loc_bus;
	 end if;
    if(loc_rw_q = '0' and	loc_add = "0011001")then
	   err_sel <= "00000" & loc_bus(2 downto 0);
    elsif(loc_rw_q = '1' and	loc_add = "0011000")then
	   err_sel(2 downto 0) <= err_sel(2 downto 0) + 1;
	 end if;
    if(loc_rw_q = '0' and	loc_add = "0011010")then
	   sync_cmd <= loc_bus(7 downto 2);
	 end if;
    if(loc_rw_q = '0' and	loc_add = "0011011")then
	   sync_mask <= loc_bus(7 downto 2);
	 end if;
    if(loc_rw_q = '0' and	loc_add = "0011101")then
	   hist_sel <= loc_bus;
	 end if;
    if(loc_rw_q = '0')then
      if(loc_add(6 downto 1) = "001111")then
		  if(loc_add(0) = '1')then
	       hist_ra(15 downto 8) <= "00" & loc_bus(5 downto 0);
		  else
	       hist_ra(7 downto 0) <= loc_bus;
		  end if;
		end if;
	 elsif(loc_add = "0011100")then
	   if(hist_ra(11 downto 0) = x"deb")then
	     hist_ra(13 downto 12) <= hist_ra(13 downto 12) + 1;
	     hist_ra(11 downto 0) <= x"000";
		else
	     hist_ra(11 downto 0) <= hist_ra(11 downto 0) + 1;
		end if;
	 end if;
  end if;
end process;
process(loc_clk,startup,rst_dl)
begin
  if(startup = '1')then
    resets <= x"fe"; -- DCM lost lock also generates powerup, so do not clear counters
	 start_reset <= '1';
  elsif(rst_dl(15) = '1')then
    resets <= x"00";
	 start_reset <= '0';
  elsif(loc_clk'event and loc_clk = '1')then
    if(loc_rw_q = '0' and	loc_add = "0010010")then
	   resets <= loc_bus;
	   start_reset <= '1';
	 end if;
  end if;
end process;
process(ttc_clk)
begin
  if(ttc_clk'event and ttc_clk = '1')then
	 rst_dl <= rst_dl(14 downto 0) & start_reset;
  end if;
end process;
-- following codes are used to optimize GOL_negedge
rst_dcm <= resets(1) or not locked(0); -- this initiates the calibration
i_dcm_rx: dcm
   generic map(CLKOUT_PHASE_SHIFT => "VARIABLE",
-- starting phase shift after reset
					PHASE_SHIFT => -48)
	port map	
         ( CLKIN   => rx_clk_in ,
           CLKFB   => rx_clk ,
           RST     => rst_dcm ,
           DSSEN   => '0' ,
           PSINCDEC=> psinc ,
           PSEN    => psen ,
           PSCLK   => rx_clk ,
			  CLK0    => rx_clk_dcm ,
			  CLK90    => rx_clk90_dcm ,
			  PSDONE  => psdone ,
			  STATUS  => rx_status ,
			  LOCKED  => locked(1));
bufg_rx_clk: bufg port map(o => rx_clk, i => rx_clk_dcm);
bufg_rx_clk90: bufg port map(o => rx_clk90, i => rx_clk90_dcm);
GOL_negedge <= neg_edge;
process(rx_clk)
begin
  if(rx_clk'event and rx_clk = '1')then
    ttc_clk_des_q2(1) <= ttc_clk_des_q(1);
    ttc_clk_des_q(1) <= ttc_clk_des;
	 if(calib_going = '1')then
	   neg_edge <= (not ttc_clk_des_q2(0) and ttc_clk_des_q(0)) or (ttc_clk_des_q2(1) and not ttc_clk_des_q(1));
		ps_state <=  ttc_clk_des_q2 & ttc_clk_des_q & ps;
	 end if;
  end if;
end process;
-- maximum phase shift is 10ns, must use both edges of the clock
process(rx_clk90)
begin
  if(rx_clk90'event and rx_clk90 = '1')then
    ttc_clk_des_q2(0) <= ttc_clk_des_q(0);
    ttc_clk_des_q(0) <= ttc_clk_des;
  end if;
end process;
-- code to automatic align clock edge by tuning phase shift value
process(rx_clk,locked)
begin
  if(locked /= "11")then
	 psen <= '0';
	 got_psdone <= '1'; 
	 ps_cntr <= "100000";
	 calib_going <= '1';
	 psinc <= '1';
	 ps <= "0000000";
  elsif(rx_clk'event and rx_clk = '1')then
	 psen <= ps_cntr(5) and got_psdone;
	 if(psen = '1')then
		if(psinc = '1')then
		  ps <= ps + 1;
		else
		  ps <= ps - 1;
		end if;
	 end if;
	 if((ttc_clk_des_q2(1) /= ttc_clk_des_q(1)) or (ttc_clk_des_q2(0) /= ttc_clk_des_q(0)))then
	   calib_going <= '0';
	 end if;
	 if(calib_going = '0' and ps > "0110000")then
		psinc <= '0';
	 end if;
	 if(psen = '1')then
	   got_psdone <= '0';
	 else
	   got_psdone <= got_psdone or psdone;
	 end if;
	 if(ps_cntr(5) = '1' or (calib_going = '0' and ps = "0110000"))then
	   ps_cntr <= "000000";
	 else
	   ps_cntr <= ps_cntr + 1;
	 end if;
  end if;
end process;
-- I2C interface
process(loc_clk)
begin
  if(loc_clk'event and loc_clk = '1')then
    if(loc_add(6 downto 3) = x"0" and loc_add(1 downto 0) = "11" and loc_rw_q = '0' and (loc_add(2) = '1' or loc_bus(7) = '1'))then
	   i2c_toggle <= not i2c_toggle;
	 end if;
    if(loc_add = "0000011" and loc_rw_q = '0')then
	   i2c_addr <= loc_bus(4 downto 0);
	   i2c_read <= loc_bus(7);
	 end if;
	 if(loc_add = "0000111" and loc_rw_q = '0')then
	   i2c_wdata <= loc_bus;
	 end if;
  end if;
end process;
process(ttc_clk)
begin
  if(ttc_clk'event and ttc_clk = '1')then
    i2c_sync <= i2c_sync(0) & i2c_toggle;
	 i2c_start <= i2c_sync(1) xor i2c_sync(0);
    if(i2c_start = '1')then
	   i2c_busy <= '1';
    else
	   i2c_busy <= i2c_busy and not i2c_done;
	 end if;
    if(i2c_start = '1')then
	   i2c_done <= '0';
    elsif(i2c_ec = '1')then
		i2c_done <= i2c_eoc and (data_cycle or i2c_fail);
	 end if;
	 if(i2c_busy = '0')then
	   clk_div <= "00000";
	 else	
      clk_div <= clk_div + 1;
	 end if;
	 if(clk_div = "11111")then
	   i2c_ec <= '1';
	 else
	   i2c_ec <= '0';
	 end if;
  end if;
end process;
-- i2c state machine
-- since this is the only master, no STOP is generated
process(ttc_clk)
begin
  if(ttc_clk'event and ttc_clk = '1')then
    if(i2c_start = '1')then
		i2c_active <= '0';
		i2c_phase <= "000";
		data_cycle <= '0';
    elsif(i2c_ec = '1')then
	   i2c_active <= i2c_busy;
		i2c_phase(2 downto 1) <= i2c_phase(1 downto 0);
		i2c_phase(0) <= i2c_busy and (i2c_phase(2) or not i2c_active);
		data_cycle <= i2c_active and ((i2c_eoc and i2c_phase(2)) or data_cycle);
	 end if;
  end if;
end process;
-- i2c_fail resets at the beginning of every i2c access
-- if acknoledge from slave is '1', i2c_fail is set and terminates the
-- operation
process(ttc_clk)
begin
  if(ttc_clk'event and ttc_clk = '1')then
    if(i2c_ec = '1')then
	   if(i2c_busy = '1' and i2c_active = '0')then
		  i2c_fail <= '0';
		elsif(GOL_SDA = '1' and i2c_phase(1) = '1')then
		  if(i2c_cntr = "01001" or
		    (i2c_cntr = "10010" and (data_cycle = '0' or i2c_read = '0')))then
		    i2c_fail <= '1';
		  end if;
		end if;   
	 end if;
  end if;
end process;
GOL_SCL <= i2c_scl;
-- scl generation and i2c_sr shift register 
process(ttc_clk)
begin
  if(ttc_clk'event and ttc_clk = '1')then
    if(i2c_active = '0')then
      i2c_scl <= "111";
		i2c_sr <= (others => '1');
	   i2c_eoc <= '0';
		i2c_cntr <= "00000";
    elsif(i2c_ec = '1')then
	   if(i2c_addr(4 downto 3) = "00")then
	     i2c_scl(0) <= (not i2c_phase(1) and i2c_scl(0)) or i2c_phase(0) or i2c_eoc;
		end if;
	   if(i2c_addr(4 downto 3) = "01")then
	     i2c_scl(1) <= (not i2c_phase(1) and i2c_scl(1)) or i2c_phase(0) or i2c_eoc;
		end if;
	   if(i2c_addr(4 downto 3) = "10")then
	     i2c_scl(2) <= (not i2c_phase(1) and i2c_scl(2)) or i2c_phase(0) or i2c_eoc;
		end if;
	   if(i2c_phase(0) = '1' and i2c_cntr = "00000")then
		  i2c_sr(17 downto 8) <=  "0100000" & data_cycle & (i2c_read and data_cycle) & '1';
		  if(data_cycle = '0')then
		    i2c_sr(7 downto 0) <= "00000" & i2c_addr(2 downto 0);
		  elsif(i2c_read = '1')then
		    i2c_sr(7 downto 0) <= "11111111"; 
		  else
		    i2c_sr(7 downto 0) <= i2c_wdata(7 downto 0); 
		  end if;
		elsif(i2c_phase(2) = '1')then
		  i2c_sr <= i2c_sr(16 downto 0) & '1';
		end if;
      if(i2c_phase(1) = '1' and (i2c_cntr(4) = '0' or i2c_cntr(1) = '0'))then
	     i2c_data <= i2c_data(6 downto 0) & GOL_SDA;
	   end if;
	   if(i2c_phase(2) = '1')then
	     if(i2c_cntr = "10010" or i2c_fail = '1')then
		    i2c_eoc <= '1';
		  else
		    i2c_eoc <= '0';
		  end if;
		  if(i2c_eoc = '1')then
		    i2c_cntr <= "00000";
		  else
		    i2c_cntr <= i2c_cntr + 1;
		  end if;
		end if;
	 end if;
  end if;
end process;
-- sda and scl outputs
GOL_SDA <= '0' when i2c_eoc = '1' or i2c_sr(17) = '0' else 'Z';
-- FLASH programming code
loc_ctrl(2) <= '1';
loc_ctrl(1) <= not rd_buffer;
loc_ctrl(0) <= not flash_ce;
process(loc_clk)
begin
  if(loc_clk'event and loc_clk = '1')then
    if(loc_add = "0010000" and loc_rw_q = '0' and loc_bus(0) = '1')then
	   flash_addra(7 downto 0) <= x"00";
	 elsif(loc_add = "0010001" and loc_rw_q = '0')then
	   flash_addra(7 downto 0) <= flash_addra(7 downto 0) + 1;
	 end if;
    if(loc_add = "0010000" and loc_rw_q = '0')then
	   cmd <= cmd xor loc_bus;
	 end if;
  end if;
end process;
process(loc_clk)
begin
  if(loc_clk'event and loc_clk = '1')then
    if(rd_done(1) = '1')then
	   rd_buffer <= '0';
    elsif(loc_add = "1100010" and loc_rw_q = '0' and empty_buffer = '0')then
	   rd_buffer <= '1';
	 end if;	
    if(flash_addrb(8) = '1')then
	   empty_buffer <= '1';
    elsif(loc_add = "0010001" and loc_rw_q = '0')then
	   empty_buffer <= '0';
	 end if;	
  end if;
end process;
process(flash_clk, rd_buffer)
begin
  if(rd_buffer = '0')then
	 ce_flash_addrb <= '0';
	 flash_ce <= '0';
	 flash_addrb(8 downto 0) <= "000000000";
	 rd_done <= "00";
  elsif(flash_clk'event and flash_clk = '1')then
	 ce_flash_addrb <= not ce_flash_addrb and not flash_addrb(8);
	 flash_ce <= ce_flash_addrb;
	 if(ce_flash_addrb = '1')then
	   flash_addrb(8 downto 0) <= flash_addrb(8 downto 0) + 1;
	 end if;
	 rd_done <= rd_done(0) & flash_addrb(8);
  end if;
end process;
flash_addra(10 downto 8) <= "000";
flash_addrb(10 downto 9) <= "00";
we_flash <= '1' when loc_add = "0010001" and loc_rw_q = '0' else '0';
i_flash_buf: ram256x8 PORT MAP(
		wclk => loc_clk,
		rclk => flash_clk,
		we => we_flash,
		re => ce_flash_addrb,
		di => loc_bus,
		wa => flash_addra(7 downto 0),
		ra => flash_addrb(7 downto 0),
		do => flash_buf
	);
--i_flash_buf: RAMB16_S9_S9
--		   port map(
--			   dia => loc_bus,
--				dib => x"00",
--				dipa => "0",
--				dipb => "0",
--				ena => '1',
--				enb => ce_flash_addrb,
--				wea => we_flash,
--				web => '0',
--				ssra => '0',
--				ssrb => '0',
--				clka => loc_clk,
--				clkb => flash_clk,
--				addra => flash_addra,
--				addrb => flash_addrb,
--				dob => flash_buf
--				);
-- output monitoring
process(ttc_clk)
begin
  if(ttc_clk'event and ttc_clk = '1')then
    cmd1_sync <= cmd1_sync(0) & cmd(1);
--	 if(cmd1_sync(1) /= cmd1_sync(0) or (sync_active0_dl = '0' and sync_active(0) = '1'))then
	 if(cmd1_sync(1) /= cmd1_sync(0))then
	   we_gol2 <= '1';
	 elsif(gol2_addrb(7 downto 0) = x"ff")then
	   we_gol2 <= '0';
	 end if;
    if(we_gol2 = '0')then
	   gol2_addrb <= (others => '0');
	 else
	   gol2_addrb(7 downto 0) <= gol2_addrb(7 downto 0) + 1;
	 end if;
  end if;
end process;
process(ttc_clk2x)
begin
  if(ttc_clk2x'event and ttc_clk2x = '1')then
	 we_gol <= we_gol2;
	 sel_gol1 <= not sel_gol1 and we_gol2;
    if(sel_gol1 = '0')then
	   gol_datap <= gol0_data_i;
	 else
	   gol_datap <= gol1_data_i;
	 end if;
	 gol_addrb <= gol2_addrb(7 downto 0) & sel_gol1;
  end if;
end process;
loc_clkn <= not loc_clk;
i_gol: RAMB16_S9_S36
		   port map(
				dia => x"00",
			   dib => GOL_datap,
				dipa => "0",
				dipb => x"0",
				ena => '1',
				enb => '1',
				wea => '0',
				web => we_gol,
				ssra => '0',
				ssrb => '0',
				clka => loc_clkn,
				clkb => ttc_clk2x,
				addra => gol_addra,
				addrb => gol_addrb,
				doa => gol_buf
				);
gol_addra(0) <= gol_spy_addr(10);
gol_addra(10 downto 1) <= gol_spy_addr(9 downto 0);
i_gol2: RAMB16_S9_S36
		   port map(
				dia => x"00",
			   dib => GOL2_data_i,
				dipa => "0",
				dipb => x"0",
				ena => '1',
				enb => '1',
				wea => '0',
				web => we_gol2,
				ssra => '0',
				ssrb => '0',
				clka => loc_clkn,
				clkb => ttc_clk,
				addra => gol_spy_addr(10 downto 0),
				addrb => gol2_addrb,
				doa => gol2_buf
				);
gol_spy_data <= gol_buf when gol_spy_addr(11) = '0' else gol2_buf;
-- TP_a(33) is BC0, TP_a(31 downto 24) are BCN(7 downto 0) -- 9/11/2007
reset_senderN <= not resets(7);
check_ena <= csr1(0) and not csr1(4);
check_data_ena <= csr1(1) and not csr1(4);
test_rand_ena <= csr1(2) and not csr1(4);
process(ttc_clk)
begin
  if(ttc_clk'event and ttc_clk = '1')then
	 if(csr1(4) = '0')then
      test_datap <= test_data;
	 elsif(test_datap = x"ffffffff")then
      test_datap <= x"00000000";
	 else	
	   test_datap <= test_datap(30 downto 0) & not(test_datap(31) xor test_datap(21) xor test_datap(1) xor test_datap(0));
    end if;
  end if;
end process;
check_data <= TP_a(30 downto 24) & TP_a(33);
i_rpc_sender0: rpc_sender PORT MAP(
		resetN => reset_senderN,
		clock => ttc_clk,
		part_ena => '0',
		check_ena => '1',
		check_data => check_data,
		check_data_ena => '1',
		in_data => data(23 downto 0),
		out_data => GOL0_data_i,
		test_data => (others => '0'),
		test_ena => '0',
		test_rand_ena => '0'
	);
i_rpc_sender1: rpc_sender PORT MAP(
		resetN => reset_senderN,
		clock => ttc_clk,
		part_ena => '0',
		check_ena => '1',
		check_data => check_data,
		check_data_ena => '1',
		in_data => data(47 downto 24),
		out_data => GOL1_data_i,
		test_data => (others => '0'),
		test_ena => '0',
		test_rand_ena => '0'
	);
i_rpc_sender2: rpc_sender PORT MAP(
		resetN => reset_senderN,
		clock => ttc_clk,
		part_ena => '0',
		check_ena => '1',
		check_data => check_data,
		check_data_ena => '1',
		in_data => data(71 downto 48),
		out_data => GOL2_data_i,
		test_data => (others => '0'),
		test_ena => '0',
		test_rand_ena => '0'
	);
-- error counters
test_ena <= csr1(4) or csr1(3) or sync_active(1);
process(qpll_err)
begin
  if(qpll_err'event and qpll_err = '1')then
    qpll_err_toggle <= not qpll_err_toggle;
  end if;
end process;
process(ttc_clk)
begin
  if(ttc_clk'event and ttc_clk = '1')then
	 err_cntr_wa <= err_cntr_wa + 1;
	 if(err_cntr_wa = "101")then
	   DCM_lock_sync <= DCM_lock_sync(0) & locked(1);
		gol0_ready_sync <= gol0_ready_sync(0) & GOL_ready(0);
		gol1_ready_sync <= gol1_ready_sync(0) & GOL_ready(1);
		gol2_ready_sync <= gol2_ready_sync(0) & GOL_ready(2);
		qpll_lock_sync <= qpll_lock_sync(0) & qpll_locked;
		qpll_err_sync <= qpll_err_sync(0) & qpll_err_toggle;
	 end if;
    case err_cntr_wa is
      when "000" => err <= gol0_ready_sync(1) and not gol0_ready_sync(0);
      when "001" => err <= gol1_ready_sync(1) and not gol1_ready_sync(0);
      when "010" => err <= gol2_ready_sync(1) and not gol2_ready_sync(0);
      when "011" => err <= qpll_lock_sync(1) and not qpll_lock_sync(0);
      when "100" => err <= qpll_err_sync(1) xor qpll_err_sync(0);
      when "101" => err <= bc0_err;
      when "110" => err <= bcn_mismatch;
      when others => err <= DCM_lock_sync(1) and not DCM_lock_sync(0);
    end case;
  end if;
end process;
g_err_cntr: for i in 0 to 7 generate
  i_err_cntr : RAM16X1D
   port map (
      DPO => err_cntr(i),     -- Read-only 1-bit data output for DPRA
      SPO => err_cntrp(i),     -- R/W 1-bit data output
      A0 => err_cntr_wa(0),       -- R/W address[0] input bit
      A1 => err_cntr_wa(1),       -- R/W address[1] input bit
      A2 => err_cntr_wa(2),       -- R/W address[2] input bit
      A3 => '0',       -- R/W ddress[3] input bit
      D => d_err_cntr(i),         -- Write 1-bit data input
      DPRA0 => err_sel(0), -- Read-only address[0] input bit
      DPRA1 => err_sel(1), -- Read-only address[1] input bit
      DPRA2 => err_sel(2), -- Read-only address[2] input bit
      DPRA3 => '0', -- Read-only address[3] input bit
      WCLK => ttc_clk,   -- Write clock input
      WE => we_err_cntr        -- Write enable input
   );
end generate;
d_err_cntr <= x"00" when resets(0) = '1' else err_cntrp + 1;
we_err_cntr <= '1' when resets(0) = '1' or (err = '1' and err_cntrp /= x"ff") else '0';
-- histogramming, four TP_x at a time
process(ttc_clk)
begin
  if(ttc_clk'event and ttc_clk = '1')then
	 bc0_a <= TP_a(33);
	 bc0_b <= TP_b(33);
	 if(bcn = x"dea")then
	   bc0 <= '1';
	 else
	   bc0 <= '0';
	 end if;
	 if(bc0_a = '1' or bc0 = '1')then
	   bcn <= x"000";
	 else
	   bcn <= bcn + 1;
	 end if;
	 if(bc0_a /= bc0)then
	   bc0_err <= '1';
    elsif(err_cntr_wa = "101")then
	   bc0_err <= '0';
	 end if;
	 if(err_cntr_wa(2) = '0')then
	   bc_errp <= '0';
	 elsif(err_cntr_wa(1) = '1' and err_cntrp /= x"00")then
	   bc_errp <= '1';
	 end if;
	 if(err_cntr_wa = "000")then
	   bc_err <= bc_errp;
	 end if;
	 if(bc0_a /= bc0_b)then
	   bcn_mismatch <= '1';
    elsif(err_cntr_wa = "110")then
	   bcn_mismatch <= '0';
	 end if;
	 if(bc0 = '1')then
	   rst_hist(2 downto 1) <= rst_hist(1 downto 0);
	   if(rst_hist(2) /= rst_hist(1))then
	     rst_hist(3) <= '1';
	   else
	     rst_hist(3) <= '0';
	   end if;
	 end if;
	 if(rst_hist(3) = '1')then
	   TP_sel_q <= x"00";
	 else
	   if(sim_hist(3) = '0')then
		  TP_sel_q(7 downto 4) <= TP_sel;
	   elsif(orbit_cntr > bcnp)then
	     TP_sel_q(7 downto 4) <= x"0";
	   else
	     TP_sel_q(7 downto 4) <= hist_sel(7 downto 4);
	   end if;
	   TP_sel_q(3 downto 0) <= TP_sel_q(7 downto 4);
	 end if;
	 if(bcn = x"dea")then
	   bcnp <= hist_sel(1 downto 0) & "00";
	 else
	   bcnp <= bcnp + 1;
	 end if;
	 if(bcn = x"dea")then
	   sim_hist(2 downto 1) <= sim_hist(1 downto 0);
	   if(sim_hist(2) /= sim_hist(1))then
	     sim_hist(3) <= '1';
	   elsif(orbit_cntr = x"f")then
	     sim_hist(3) <= '0';
	   end if;
		if(sim_hist(3) = '0')then
		  orbit_cntr <= (others => '0');
		else
		  orbit_cntr <= orbit_cntr + 1;
		end if;
	 end if;
  end if;
end process;
process(hist_sel,TP_a,TP_b)
begin
  case hist_sel(3 downto 0) is
    when x"0" => TP_sel <= mux_in_a(3 downto 0);
    when x"1" => TP_sel <= mux_in_a(7 downto 4);
    when x"2" => TP_sel <= mux_in_a(11 downto 8);
    when x"3" => TP_sel <= mux_in_a(15 downto 12);
    when x"4" => TP_sel <= mux_in_a(19 downto 16);
    when x"5" => TP_sel <= mux_in_a(23 downto 20);
    when x"6" => TP_sel <= mux_in_a(27 downto 24);
    when x"7" => TP_sel <= mux_in_a(31 downto 28);
    when x"8" => TP_sel <= mux_in_b(3 downto 0);
    when x"9" => TP_sel <= mux_in_b(7 downto 4);
    when x"a" => TP_sel <= mux_in_b(11 downto 8);
    when x"b" => TP_sel <= mux_in_b(15 downto 12);
	 when others => TP_sel <= x"0";
  end case;
end process;
g_hist: for i in 0 to 7 generate
   i_hist : RAMB16_S9_S9
   port map (
      DOA => hist(i*8+7 downto i*8),      -- Port A 8-bit Data Output
      DOB => hist_do(i*8+7 downto i*8),      -- Port B 8-bit Data Output
      ADDRA => hist_ra(11 downto 1),  -- Port A 11-bit Address Input
      ADDRB => bcn(11 downto 1),  -- Port B 11-bit Address Input
      CLKA => loc_clkn,    -- Port A Clock
      CLKB => ttc_clk,    -- Port B Clock
      DIA => x"00",      -- Port A 8-bit Data Input
      DIPA => "0",    -- Port-A 1-bit parity Input
      DIB => sum(i*9+7 downto i*9),      -- Port B 8-bit Data Input
      DIPB => "0",    -- Port-B 1-bit parity Input
      ENA => '1',      -- Port A RAM Enable Input
      ENB => '1',      -- PortB RAM Enable Input
      SSRA => rst_hist(3),    -- Port A Synchronous Set/Reset Input
      SSRB => rst_hist(3),    -- Port B Synchronous Set/Reset Input
      WEA => '0',      -- Port A Write Enable Input
      WEB => we_hist(i)       -- Port B Write Enable Input
   );
  sum(i*9+8 downto i*9) <= ('0' & hist_do(i*8+7 downto i*8)) + (x"00" & TP_sel_q(i));
  we_hist(i) <= bcn(0) and not sum(i*9+8);
end generate;
process(hist,hist_ra)
variable rd_sel : std_logic_vector(2 downto 0);
begin
  rd_sel := hist_ra(0) & hist_ra(13 downto 12);
  case rd_sel is
    when "000" => hist_data <= hist(7 downto 0);
    when "001" => hist_data <= hist(15 downto 8);
    when "010" => hist_data <= hist(23 downto 16);
    when "011" => hist_data <= hist(31 downto 24);
    when "100" => hist_data <= hist(39 downto 32);
    when "101" => hist_data <= hist(47 downto 40);
    when "110" => hist_data <= hist(55 downto 48);
    when others => hist_data <= hist(63 downto 56);
  end case;
end process;
end Behavioral;
